// Runtime names and paths

#define _GNU_SOURCE

#include "runtime.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static _Bool init_failed;
static char *etc_dir;
static char *root_dir;
static char *run_dir;
static char *share_dir;
static char *var_lib_dir;

static _Bool set_root_dir(_Bool *usr_dir);

__attribute__((constructor)) static void unit_constructor(void) {
	_Bool usr_dir;
	if (!set_root_dir(&usr_dir))
		goto error;

	const char *root_str = root_dir? : "";
	const char *prog_name = program_invocation_short_name;

	if (asprintf(&etc_dir, "%s/etc/%s", root_str, prog_name) == -1) {
		etc_dir = NULL;
		goto error;
	}

	if (asprintf(&run_dir, "%s/var/run/%s", root_str, prog_name) == -1) {
		run_dir = NULL;
		goto error;
	}

	if (asprintf(&share_dir, "%s%s/share/%s", root_str, usr_dir? "/usr": "",
		prog_name) == -1) {
		share_dir = NULL;
		goto error;
	}

	if (asprintf(&var_lib_dir, "%s/var/lib/%s", root_str, prog_name)
		== -1) {
		var_lib_dir = NULL;
		goto error;
	}

	return;

error:
	init_failed = 1;
}

__attribute__((destructor)) static void unit_destructor(void) {
	free(etc_dir);
	free(root_dir);
	free(run_dir);
	free(share_dir);
	free(var_lib_dir);
}

static char *remove_std_dirs(char *path, char *end) {
	size_t path_len = end - path;

	static const char *dirs[] = { "/bin", "/games" };
	int i;
	for (i = 0; i < 2; i++) {
		const char *dir = dirs[i];
		if (path_len >= strlen(dir)) {
			char *p = end - strlen(dir);
			if (!strcmp(p, dir)) {
				*p = '\0';
				return p;
			}
		}
	}

	// TODO: Ignore prog_name
	const char *prog_name = program_invocation_short_name;

	// TODO: Check multiarch dir
	static const char LIB[] = "/lib/";
	if (path_len >= strlen(LIB) + strlen(prog_name)) {
		char *p = end - strlen(prog_name);
		if (!strcmp(p, prog_name)) {
			p -= strlen(LIB);
			if (!strncmp(p, LIB, strlen(LIB))) {
				*p = '\0';
				return p;
			}
		}
	}
	return NULL;
}

const char *runtime_etc_dir(void) {
	return etc_dir;
}

_Bool runtime_init_failed(void) {
	return init_failed;
}

const char *runtime_prog_name(void) {
	return program_invocation_short_name;
}

const char *runtime_run_dir(void) {
	return run_dir;
}

const char *runtime_share_dir(void) {
	return share_dir;
}

const char *runtime_var_lib_dir(void) {
	return var_lib_dir;
}

static _Bool set_root_dir(_Bool *usr_dir) {
	char *path;
	if (program_invocation_name[0] == '/')
		path = strdup(program_invocation_name);
	else if (strchr(program_invocation_name, '/')) {
		char *dir = get_current_dir_name();
		if (!dir)
			return 0;
		if (asprintf(&path, "%s/%s", dir, program_invocation_name)
			== -1)
			path = NULL;
		free(dir);
	}
	else
		path = canonicalize_file_name("/proc/self/exe");

	if (!path)
		return 0;

	char *slash = strrchr(path, '/');
	if (!slash)
		goto error;
	*slash = '\0';

	_Bool this_usr_dir = 0;

	char *new_end = remove_std_dirs(path, slash);
	if (new_end) {
		size_t path_len = new_end - path;
		static const char USR[] = "/usr";
		if (path_len >= strlen(USR)) {
			char *p = new_end - strlen(USR);
			if (!strcmp(p, USR)) {
				*p = '\0';
				this_usr_dir = 1;
			}
		}
	}

	if (path[0]) {
		root_dir = strdup(path);
		if (!root_dir)
			goto error;
	}

	*usr_dir = this_usr_dir;

	free(path);
	return 1;

error:
	free(path);
	return 0;
}
